"""Reasoner API models."""

import copy
import hashlib

from typing import Optional, Callable, Union


from .results import Results, Analysis
from .qgraph import QueryGraph, PathfinderQueryGraph
from pydantic import (
    AnyHttpUrl,
    ValidationInfo,
    ConfigDict,
    Field,
    model_validator,
)

from .base_model import BaseModel
from .utils import HashableSequence
from .kgraph import Edge, KnowledgeGraph
from .shared import EdgeIdentifier, LogEntry, LogLevel
from .workflow import Workflow
from .auxgraphs import AuxiliaryGraphs
from typing import Annotated


class Message(BaseModel):
    """Message."""

    query_graph: Annotated[
        Optional[Union[QueryGraph, PathfinderQueryGraph]],
        Field(
            title="query graph",
        ),
    ] = None
    # query_graph: Optional[QueryGraph] = None
    knowledge_graph: Annotated[
        Optional[KnowledgeGraph],
        Field(
            title="knowledge graph",
        ),
    ] = None
    results: Annotated[
        Optional[Results],
        Field(
            title="list of results",
        ),
    ] = None
    auxiliary_graphs: Annotated[
        Optional[AuxiliaryGraphs],
        Field(
            title="dict of auxiliary graphs",
        ),
    ] = None
    model_config = ConfigDict(title="message", extra="forbid")

    def update(self, other: object, normalize: bool = True) -> None:
        """Updates one message with information from another.
        Can run with normalize=false, if both messages are normalized already."""
        if not isinstance(other, Message):
            raise TypeError("Message may only be updated with another Message.")

        if hash(self.query_graph) != hash(other.query_graph):
            raise NotImplementedError("Query graph merging not supported yet")
        # Make a copy because normalization will modify results
        other = other.model_copy(deep=True)

        if other.knowledge_graph:
            if not self.knowledge_graph:
                self.knowledge_graph = KnowledgeGraph()
            # Normalize edges of incoming KG
            # This will place KG edge keys into the same hashing system
            # So that equivalence is determined by hash collision
            if normalize:
                other._normalize_kg_edge_ids()
            # The knowledge graph can now be updated because edge keys will be
            # hashed using the same method. The knowledge graph update method
            # will handle concatenating properties when necessary.
            self.knowledge_graph.update(other.knowledge_graph)
        if other.results:
            if self.results:
                self.results.update(other.results)
            else:
                self.results = other.results
        if other.auxiliary_graphs:
            if self.auxiliary_graphs:
                self.auxiliary_graphs.update(other.auxiliary_graphs)
            else:
                self.auxiliary_graphs = other.auxiliary_graphs

    @model_validator(mode="after")
    def normalize_on_parse(self, info: ValidationInfo) -> "Message":
        normalize = True
        if isinstance(info.context, dict) and not info.context.get("normalize", True):
            normalize = False
        if normalize and self.knowledge_graph is not None:
            self._normalize_kg_edge_ids()
        return self

    def normalize(self) -> None:
        self._normalize_kg_edge_ids()

    def _normalize_kg_edge_ids(self) -> None:
        """
        Replace edge IDs with a hash of the edge object
        """
        self._update_kg_edge_ids(
            lambda edge: EdgeIdentifier(
                hashlib.blake2b(
                    hash(edge).to_bytes(16, byteorder="big", signed=True), digest_size=6
                ).hexdigest()
            )
        )

    def _update_kg_edge_ids(self, update_func: Callable[[Edge], EdgeIdentifier]):
        """
        Replace edge IDs using the specified function
        """

        if self.knowledge_graph is None:
            return

        # Mapping of old to new edge IDs
        edge_id_mapping: dict[EdgeIdentifier, EdgeIdentifier] = {}

        # Make a copy of the edge keys because we're about to change them
        for edge_id in list(self.knowledge_graph.edges.keys()):
            edge = self.knowledge_graph.edges.pop(edge_id)
            new_edge_id = update_func(edge)

            edge_id_mapping[edge_id] = new_edge_id

            # Update knowledge graph
            self.knowledge_graph.edges[new_edge_id] = edge

        # Update auxiliary graphs
        if self.auxiliary_graphs:
            aux_len = len(self.auxiliary_graphs.values())
            aux_num = 0
            for auxiliary_graph in self.auxiliary_graphs.values():
                aux_num += 1
                if auxiliary_graph.edges:
                    edges_len = len(auxiliary_graph.edges)
                    edges_num = 0
                    aux_edges = copy.deepcopy(auxiliary_graph.edges)
                    for aux_edge in aux_edges:
                        edges_num += 1
                        auxiliary_graph.edges.discard(aux_edge)
                        try:
                            new_edge_id = edge_id_mapping[aux_edge]
                            auxiliary_graph.edges.add(new_edge_id)
                            edge_id_mapping[new_edge_id] = new_edge_id
                        except KeyError:
                            raise Exception(
                                f"Aux graph edge id {aux_edge} not found in edge id mapping"
                            )
                    if edges_len != edges_num:
                        raise Exception("Missed an aux graph edge normalization.")
                else:
                    raise Exception("This aux graph has no edges")
            if aux_len != aux_num:
                raise Exception("Missed an aux graph normalization.")

        # Update results
        if self.results:
            for result in self.results:
                if result and result.analyses:
                    for analysis in result.analyses:
                        if isinstance(analysis, Analysis):
                            for edge_binding_list in analysis.edge_bindings.values():
                                for eb in edge_binding_list:
                                    eb.id = edge_id_mapping[eb.id]


class Query(BaseModel):
    """Request."""

    message: Message
    log_level: Optional[LogLevel] = None
    workflow: Optional[Workflow] = None
    bypass_cache: Optional[bool] = False
    model_config = ConfigDict(
        title="query", extra="allow", json_schema_extra={"x-body-name": "request_body"}
    )


class AsyncQuery(BaseModel):
    """AsyncQuery."""

    callback: AnyHttpUrl
    message: Message
    log_level: Optional[LogLevel] = None
    workflow: Optional[Workflow] = None
    bypass_cache: Optional[bool] = False
    model_config = ConfigDict(
        title="query", extra="allow", json_schema_extra={"x-body-name": "request_body"}
    )


class Response(BaseModel):
    """Response."""

    message: Message
    logs: HashableSequence[LogEntry] = Field(
        default_factory=lambda: HashableSequence[LogEntry]()
    )
    status: Optional[str] = None
    description: Optional[str] = None
    workflow: Optional[Workflow] = None
    schema_version: Optional[str] = None
    biolink_version: Optional[str] = None

    model_config = ConfigDict(title="response", extra="allow")

    @model_validator(mode="after")
    def normalize(self, info: ValidationInfo) -> "Response":
        normalize = True
        if isinstance(info.context, dict) and not info.context.get("normalize", True):
            normalize = False
        if normalize and self.message.knowledge_graph is not None:
            self.message.normalize()
        return self


class AsyncQueryResponse(BaseModel):
    """ "Async Query Response."""

    status: Optional[str] = None
    description: Optional[str] = None
    job_id: Annotated[str, Field(title="job id")]
    model_config = ConfigDict(title="async query response", extra="allow")


class AsyncQueryStatusResponse(BaseModel):
    """Async Query Status Response."""

    status: str
    description: str
    logs: HashableSequence[LogEntry]
    response_url: Optional[str] = None

    model_config = ConfigDict(title="async query status response", extra="allow")



"""Auxiliary Graphs model"""

from typing import Annotated, Any
from pydantic import ConfigDict, Field

from .base_model import BaseModel
from .utils import HashableMapping, HashableSet
from .shared import Attribute, EdgeIdentifier


class AuxiliaryGraph(BaseModel):
    """Auxiliary Graph"""

    edges: Annotated[
        HashableSet[EdgeIdentifier], Field(title="edges in auxiliary graph")
    ]
    attributes: HashableSet[Attribute]

    model_config = ConfigDict(title="auxiliary graph", extra="allow")


class AuxiliaryGraphs(HashableMapping[str, AuxiliaryGraph]):
    """Auxiliary Graphs"""

    model_config = ConfigDict(title="auxiliary graphs")

    def update(self, other: object, *_args: Any, **_kwargs: Any):
        if not isinstance(other, HashableMapping):
            raise TypeError("AuxiliaryGraphs may only be updated with AuxiliaryGraphs.")
        self.root.update(other.root)

    def values(self):
        return self.root.values()

    def items(self):
        return self.root.items()

    def keys(self):
        return self.root.keys()

    def __setitem__(self, k: str, v: AuxiliaryGraph):
        self.root[k] = v

    def __getitem__(self, k: str):
        return self.root[k]

    def __iter__(self):
        return iter(self.root)



"""Results models."""

import copy
from typing import Annotated, Optional, Union

from pydantic import ConfigDict, Field, model_validator

from .base_model import BaseModel
from .utils import HashableMapping, HashableSet, HashableSequence
from .shared import Attribute, CURIE, EdgeIdentifier


class EdgeBinding(BaseModel):
    """Edge binding."""

    id: Annotated[EdgeIdentifier, Field(title="knowledge graph id")]
    attributes: HashableSet[Attribute]

    model_config = ConfigDict(
        title="edge binding",
        json_schema_extra={
            "example": {
                "id": "string",
            },
        },
        extra="allow",
    )


class NodeBinding(BaseModel):
    """Node binding."""

    id: Annotated[
        CURIE,
        Field(
            title="knowledge graph id",
        ),
    ]
    query_id: Annotated[Optional[CURIE], Field(title="query graph id")] = None
    attributes: HashableSet[Attribute]

    model_config = ConfigDict(
        title="node binding",
        json_schema_extra={
            "example": {
                "id": "x:string",
            },
        },
        extra="allow",
    )


class PathBinding(BaseModel):
    """Path binding."""

    id: Annotated[str, Field(title="auxiliary graph id")]

    model_config = ConfigDict(
        title="path binding",
        json_schema_extra={"example": {"id": "string"}},
        extra="allow",
    )


class BaseAnalysis(BaseModel):
    """Base Analysis."""

    resource_id: Annotated[
        CURIE,
        Field(
            title="resource infores",
        ),
    ]

    score: Optional[float] = None
    support_graphs: Optional[HashableSet[str]] = None
    scoring_method: Optional[str] = None
    attributes: Optional[HashableSet[Attribute]] = None


class Analysis(BaseAnalysis):
    """Standard analysis"""

    edge_bindings: Annotated[
        HashableMapping[str, HashableSet[EdgeBinding]],
        Field(
            title="list of edge bindings",
        ),
    ]

    model_config = ConfigDict(title="standard analysis", extra="allow")

    def __hash__(self) -> int:
        return hash(
            (
                self.resource_id,
                self.edge_bindings,
                self.score,
                self.support_graphs,
                self.scoring_method,
            )
        )

    def update(self, other: object):
        if not isinstance(other, Analysis):
            raise TypeError("Analysis may only be updated with another Analysis.")
        for k in other.edge_bindings:
            if k in self.edge_bindings:
                self.edge_bindings[k].update(copy.deepcopy(other.edge_bindings[k]))
            else:
                self.edge_bindings[k] = copy.deepcopy(other.edge_bindings[k])
        if other.attributes:
            if self.attributes:
                self.attributes.update(other.attributes)
            else:
                self.attributes = other.attributes
        if other.support_graphs:
            if self.support_graphs:
                self.support_graphs.update(other.support_graphs)
            else:
                self.support_graphs = other.support_graphs


class PathfinderAnalysis(BaseAnalysis):
    """Pathfinder Analysis."""

    path_bindings: Annotated[
        HashableMapping[str, HashableSet[PathBinding]],
        Field(
            title="list of path bindings",
        ),
    ]

    model_config = ConfigDict(title="pathfinder analysis", extra="allow")

    def __hash__(self) -> int:
        return hash(
            (
                self.resource_id,
                self.path_bindings,
                self.score,
                self.support_graphs,
                self.scoring_method,
            )
        )

    def update(self, other):
        for k in other.path_bindings:
            if k in self.path_bindings:
                self.path_bindings[k].update(copy.deepcopy(other.path_bindings[k]))
            else:
                self.path_bindings[k] = copy.deepcopy(other.path_bindings[k])
        if other.attributes:
            if self.attributes:
                self.attributes.update(other.attributes)
            else:
                self.attributes = other.attributes
        if other.support_graphs:
            if self.support_graphs:
                self.support_graphs.update(other.support_graphs)
            else:
                self.support_graphs = other.support_graphs


class Result(BaseModel):
    """Result."""

    node_bindings: Annotated[
        HashableMapping[str, HashableSet[NodeBinding]],
        Field(
            title="list of node bindings",
        ),
    ]
    analyses: Annotated[
        Union[HashableSet[Analysis], HashableSet[PathfinderAnalysis]],
        Field(
            title="list of anlysis blocks",
        ),
    ]

    model_config = ConfigDict(title="result", extra="allow")

    def update(self, other: object):
        if not isinstance(other, Result):
            raise TypeError("Result may only be updated with another Result.")
        if other.analyses:
            if self.analyses:
                for analysis in other.analyses:
                    check = True
                    for ana in self.analyses:
                        if analysis == ana:
                            ana.update(analysis)
                            check = False
                    if check:
                        self.analyses.add(analysis)
            else:
                self.analyses = other.analyses

    def __hash__(self) -> int:
        return hash(self.node_bindings)

    def combine_analyses_by_resource_id(self):
        # Useful when a service unintentionally adds multiple analyses to a single result
        # Combines all of those analyses
        combine = HashableMapping[CURIE, Analysis]()
        analyses = HashableSequence[Analysis]([analysis for analysis in self.analyses])
        for _ in range(len(analyses)):
            analysis = analyses.pop()
            if analysis.resource_id not in combine:
                combine[analysis.resource_id] = analysis
            for analysis_to_compare in analyses:
                if (
                    analysis.resource_id == analysis_to_compare.resource_id
                    and analysis != analysis_to_compare
                ):
                    combine[analysis.resource_id].update(analysis_to_compare)

        combined_analyses = HashableSet[Analysis]()
        for analysis in combine.values():
            combined_analyses.add(analysis)

        self.analyses = combined_analyses


class Results(HashableSequence[Result]):
    """Results."""

    model_config = ConfigDict(title="allow")

    def append(self, value: Result):
        self.root.append(value)

    def add(self, result: Result):
        results = self.root
        try:
            # this is slow for larger results
            results[results.index(result)].update(result)
        except ValueError:
            results.append(result)

    def __len__(self):
        return len(self.root)

    def __iter__(self):
        return self.root.__iter__()

    def __contains__(self, v):
        return self.root.__contains__(v)

    def __getitem__(self, i):
        return self.root.__getitem__(i)

    def update(self, other: object):
        results = HashableMapping[int, Result](
            {hash(result): result for result in self.root}
        )
        for result in other:
            if not isinstance(result, Result):
                result = Result.model_validate(result)
            result_hash = hash(result)
            if result_hash in results:
                results[result_hash].update(result)
            else:
                results[hash(result)] = result
        self.root.clear()
        self.root.extend(results.values())

    @model_validator(mode="after")
    def merge_results(self):
        results: dict[int, Result] = {}
        for result in self.root:
            result_hash = hash(result)
            if result_hash in results:
                results[result_hash].update(result)
            else:
                results[result_hash] = result

        self.root.clear()
        self.root.extend(results.values())
        return self



"""Knowledge graph models."""

from typing import Annotated, Any, Optional

from pydantic import ConfigDict, Field

from .shared import (
    Attribute,
    BiolinkEntity,
    BiolinkPredicate,
    CURIE,
    EdgeIdentifier,
    Qualifier,
    ResourceRoleEnum,
)
from .base_model import BaseModel
from .utils import HashableMapping, HashableSet


class Node(BaseModel):
    """Knowledge graph node."""

    categories: HashableSet[BiolinkEntity]
    name: Optional[str] = None
    attributes: HashableSet[Attribute]
    is_set: Optional[bool] = None

    model_config = ConfigDict(
        title="knowledge-graph node",
        json_schema_extra={
            "example": {
                "category": "string",
            },
        },
        extra="forbid",
        populate_by_name=True,
    )

    def update(self, other: object) -> None:
        if not isinstance(other, Node):
            raise TypeError("Node may only be updated with Node.")
        if other.name:
            self.name = other.name

        if other.categories:
            if self.categories:
                self.categories.update(other.categories)
            else:
                self.categories = other.categories

        if other.attributes:
            if self.attributes:
                self.attributes.update(other.attributes)
            else:
                self.attributes = other.attributes


class RetrievalSource(BaseModel):
    """A component of source retrieval provenance"""

    resource_id: Annotated[CURIE, Field(title="infores for source")]
    resource_role: Annotated[ResourceRoleEnum, Field(title="source type")]
    upstream_resource_ids: Optional[HashableSet[CURIE]] = None
    source_record_urls: Optional[HashableSet[str]] = None

    def __hash__(self) -> int:
        return hash((self.resource_id, self.resource_role))

    def update(self, other: object):
        if not isinstance(other, RetrievalSource):
            raise TypeError("RetrievalSource may only be updated with RetrievalSource.")
        if other.upstream_resource_ids:
            if self.upstream_resource_ids:
                self.upstream_resource_ids.update(other.upstream_resource_ids)
            else:
                self.upstream_resource_ids = other.upstream_resource_ids


class Edge(BaseModel):
    """Knowledge graph edge."""

    subject: Annotated[
        CURIE,
        Field(
            title="subject node id",
        ),
    ]
    object: Annotated[
        CURIE,
        Field(
            title="object node id",
        ),
    ]
    predicate: Annotated[BiolinkPredicate, Field(title="edge predicate")]
    sources: Annotated[
        HashableSet[RetrievalSource], Field(title="list of source retrievals")
    ]
    qualifiers: Optional[HashableSet[Qualifier]] = None
    attributes: Optional[HashableSet[Attribute]]

    model_config = ConfigDict(title="knowledge-graph edge", extra="forbid")

    def __hash__(self) -> int:
        primary_knowledge_source = self.get_primary_knowedge_source()
        return hash(
            (
                self.subject,
                self.object,
                self.predicate,
                self.qualifiers,
                primary_knowledge_source,
            )
        )

    def update(self, other: Any):
        if not isinstance(other, Edge):
            raise TypeError("Edge may only be updated with Edge.")
        if other.attributes:
            if self.attributes:
                # We need to make sure we don't add a second KL/AT
                new_attributes = HashableSet[Attribute](set())
                for attribute in other.attributes:
                    if attribute.attribute_type_id not in (
                        "biolink:knowledge_level",
                        "biolink:agent_type",
                    ):
                        new_attributes.add(attribute)
                self.attributes.update(new_attributes)
            else:
                self.attributes = other.attributes
        if other.sources:
            if self.sources:
                # We need to do the union of the two sets and then update each
                # source with the information from the other list
                self.sources.update(other.sources)
                for source in self.sources:
                    for other_source in other.sources:
                        if source == other_source:
                            source.update(other_source)
            else:
                self.sources = other.sources

    def get_primary_knowedge_source(self):
        for source in self.sources:
            if source.resource_role == "primary_knowledge_source":
                return source.resource_id


class KnowledgeGraph(BaseModel):
    """Knowledge graph."""

    nodes: HashableMapping[CURIE, Node] = Field(
        default_factory=lambda: HashableMapping[CURIE, Node]()
    )
    edges: HashableMapping[EdgeIdentifier, Edge] = Field(
        default_factory=lambda: HashableMapping[EdgeIdentifier, Edge]()
    )

    model_config = ConfigDict(title="knowledge graph", extra="allow")

    def update(self, other: object) -> None:
        if not isinstance(other, KnowledgeGraph):
            raise TypeError("KnowledgeGraph may only be updated with KnowledgeGraph.")

        for key, value in other.nodes.items():
            existing = self.nodes.get(key, None)
            if existing:
                existing.update(value)
            else:
                self.nodes[key] = value

        for key, value in other.edges.items():
            existing = self.edges.get(key, None)
            if existing:
                existing.update(value)
            else:
                self.edges[key] = value



from pydantic import AfterValidator, ConfigDict
from reasoner_pydantic.shared import CURIE, KnowledgeType
from typing import Annotated, Optional

from reasoner_pydantic import BiolinkEntity, BiolinkPredicate

from .base_model import BaseModel
from .utils import HashableMapping, HashableSequence, nonzero_validator


class MetaAttribute(BaseModel):
    """MetaAttribute."""

    attribute_type_id: CURIE
    attribute_source: Optional[str] = None
    original_attribute_names: Optional[HashableSequence[str]] = None
    constraint_use: Optional[bool] = False
    constraint_name: Optional[str] = None


class MetaNode(BaseModel):
    id_prefixes: Annotated[HashableSequence[str], AfterValidator(nonzero_validator)]
    attributes: Optional[HashableSequence[MetaAttribute]] = None
    model_config = ConfigDict(extra="forbid")


class MetaQualifier(BaseModel):
    qualifier_type_id: CURIE
    applicable_values: Optional[HashableSequence[str]] = None


class MetaEdge(BaseModel):
    subject: BiolinkEntity
    predicate: BiolinkPredicate
    object: BiolinkEntity
    qualifiers: Optional[HashableSequence[MetaQualifier]] = None
    attributes: Optional[HashableSequence[MetaAttribute]] = None
    knowledge_types: Optional[HashableSequence[KnowledgeType]] = None
    association: Optional[BiolinkEntity] = None
    model_config = ConfigDict(extra="forbid")


class MetaKnowledgeGraph(BaseModel):
    nodes: HashableMapping[str, MetaNode]
    edges: HashableSequence[MetaEdge]



"""Shared models."""

from __future__ import annotations

import re
from datetime import datetime
from enum import Enum
from typing import Annotated, Hashable, Optional

from pydantic import BeforeValidator, ConfigDict, Field
from pydantic.types import StringConstraints

from .base_model import BaseModel
from .utils import HashableSequence, make_hashable


# TODO: potential add validation for structure of CURIE
CURIE = str


class ResourceRoleEnum(str, Enum):
    """Types of resources"""

    aggregator_knowledge_source = "aggregator_knowledge_source"
    primary_knowledge_source = "primary_knowledge_source"
    supporting_data_source = "supporting_data_source"


class KnowledgeType(str, Enum):
    "Knowledge Type."

    lookup = "lookup"
    inferred = "inferred"


EdgeIdentifier = str


class Attribute(BaseModel):
    """Node/edge attribute."""

    attribute_type_id: Annotated[CURIE, Field(title="type")]
    value: Annotated[Hashable, BeforeValidator(make_hashable)]
    value_type_id: Optional[CURIE] = None
    original_attribute_name: Optional[str] = None
    value_url: Optional[str] = None
    attribute_source: Optional[str] = None
    description: Optional[str] = None
    attributes: Optional[HashableSequence["Attribute"]] = None
    model_config = ConfigDict(extra="forbid")


BiolinkQualifier = Annotated[
    str, StringConstraints(pattern=re.compile("^biolink:[a-z][a-z_]*$"))
]


class Qualifier(BaseModel):
    """Edge qualifier."""

    qualifier_type_id: Annotated[BiolinkQualifier, Field(title="type")]

    qualifier_value: Annotated[str, Field(title="value")]


BiolinkPredicate = Annotated[
    str, StringConstraints(pattern=re.compile("^biolink:[a-z][a-z_]*$"))
]
BiolinkEntity = Annotated[
    str, StringConstraints(pattern=re.compile("^biolink:[A-Z][a-zA-Z_]*$"))
]


class LogLevel(str, Enum):
    """Log level."""

    ERROR = "ERROR"
    WARNING = "WARNING"
    INFO = "INFO"
    DEBUG = "DEBUG"


class LogEntry(BaseModel):
    """Log entry."""

    timestamp: Optional[datetime] = None
    level: Optional[LogLevel] = None
    code: Optional[str] = None
    message: str = ""
    model_config = ConfigDict(extra="allow")



"""Operations models."""

from enum import Enum
from typing import Any, Literal, Optional, Union


from .base_model import BaseModel, RootModel
from .utils import HashableSequence
from .shared import BiolinkPredicate
from pydantic import Field, ConfigDict
from typing import Annotated


class RunnerAllowList(BaseModel):
    allowlist: Optional[HashableSequence[str]] = None
    timeout: Optional[float] = None
    model_config = ConfigDict(extra="forbid")


class RunnerDenyList(BaseModel):
    denylist: Optional[HashableSequence[str]] = None
    timeout: Optional[float] = None
    model_config = ConfigDict(extra="forbid")


class RunnerTimeout(BaseModel):
    timeout: Optional[float] = None
    model_config = ConfigDict(extra="forbid")


RunnerParameters = RootModel[
    Optional[Union[RunnerAllowList, RunnerDenyList, RunnerTimeout]]
]


class BaseOperation(BaseModel):
    runner_parameters: Optional[RunnerParameters] = None
    model_config = ConfigDict(extra="forbid")


class OperationAnnotate(BaseOperation):
    id: Literal["annotate"]
    parameters: Optional[Any] = None
    model_config = ConfigDict(extra="forbid")


class AnnotateEdgesParameters(BaseModel):
    attributes: Optional[HashableSequence[str]] = None


class OperationAnnotateEdges(BaseOperation):
    id: Literal["annotate_edges"]
    parameters: Optional[AnnotateEdgesParameters] = None
    model_config = ConfigDict(extra="forbid")


class AnnotateNodesParameters(BaseModel):
    attributes: Optional[HashableSequence[str]] = None


class OperationAnnotateNodes(BaseOperation):
    id: Literal["annotate_nodes"]
    parameters: Optional[AnnotateNodesParameters] = None
    model_config = ConfigDict(extra="forbid")


class OperationBind(BaseOperation):
    id: Literal["bind"]
    parameters: Optional[Any] = None
    model_config = ConfigDict(extra="forbid")


class OperationCompleteResults(BaseOperation):
    id: Literal["complete_results"]
    parameters: Optional[Any] = None
    model_config = ConfigDict(extra="forbid")


class EnrichResultsParameters(BaseModel):
    pvalue_threshold: Annotated[float, Field(ge=0.0, le=1.0)] = 1e-6
    qnode_keys: Optional[HashableSequence[str]] = None
    predicates_to_exclude: Optional[HashableSequence[BiolinkPredicate]] = None


class OperationEnrichResults(BaseOperation):
    id: Literal["enrich_results"]
    parameters: Optional[EnrichResultsParameters] = None
    model_config = ConfigDict(extra="forbid")


class FillAllowParameters(BaseModel):
    allowlist: Optional[HashableSequence[str]] = None
    qedge_keys: Optional[HashableSequence[str]] = None
    model_config = ConfigDict(extra="forbid")


class FillDenyParameters(BaseModel):
    denylist: Optional[HashableSequence[str]] = None
    qedge_keys: Optional[HashableSequence[str]] = None
    model_config = ConfigDict(extra="forbid")


FillParameters = RootModel[Union[FillAllowParameters, FillDenyParameters]]


class OperationFill(BaseOperation):
    id: Literal["fill"]
    parameters: Optional[FillParameters] = None
    model_config = ConfigDict(extra="forbid")


class OperationFilterResults(BaseOperation):
    id: Literal["filter_results"]
    parameters: Optional[Any] = None
    model_config = ConfigDict(extra="forbid")


class FilterResultsTopNParameters(BaseModel):
    max_results: Annotated[int, Field(ge=0)]
    model_config = ConfigDict(extra="forbid")


class OperationFilterResultsTopN(BaseOperation):
    id: Literal["filter_results_top_n"]
    parameters: Optional[FilterResultsTopNParameters] = None
    model_config = ConfigDict(extra="forbid")


class OperationFilterKgraph(BaseOperation):
    id: Literal["filter_kgraph"]
    parameters: Optional[Any] = None
    model_config = ConfigDict(extra="forbid")


class AboveOrBelowEnum(str, Enum):
    """'above' or 'below'."""

    above = "above"
    below = "below"


class FilterKgraphContinuousKedgeAttributeParameters(BaseModel):
    edge_attribute: str
    threshold: float
    remove_above_or_below: AboveOrBelowEnum
    qedge_keys: Optional[HashableSequence[str]] = None
    qnode_keys: HashableSequence[str] = Field(
        default_factory=lambda: HashableSequence[str]()
    )


class OperationFilterKgraphContinuousKedgeAttribute(BaseOperation):
    id: Literal["filter_kgraph_continuous_kedge_attribute"]
    parameters: FilterKgraphContinuousKedgeAttributeParameters
    model_config = ConfigDict(extra="forbid")


class FilterKgraphDiscreteKedgeAttributeParameters(BaseModel):
    edge_attribute: str
    remove_value: Any = None
    qedge_keys: Optional[HashableSequence[str]] = None
    qnode_keys: HashableSequence[str] = Field(
        default_factory=lambda: HashableSequence[str]()
    )


class OperationFilterKgraphDiscreteKedgeAttribute(BaseOperation):
    id: Literal["filter_kgraph_discrete_kedge_attribute"]
    parameters: FilterKgraphDiscreteKedgeAttributeParameters
    model_config = ConfigDict(extra="forbid")


class FilterKgraphDiscreteKnodeAttributeParameters(BaseModel):
    node_attribute: str
    remove_value: Any = None
    qnode_keys: Optional[HashableSequence[str]] = None


class OperationFilterKgraphDiscreteKnodeAttribute(BaseOperation):
    id: Literal["filter_kgraph_discrete_knode_attribute"]
    parameters: FilterKgraphDiscreteKnodeAttributeParameters
    model_config = ConfigDict(extra="forbid")


class OperationFilterKgraphOrphans(BaseOperation):
    id: Literal["filter_kgraph_orphans"]
    parameters: Optional[Any] = None
    model_config = ConfigDict(extra="forbid")


class TopOrBottomEnum(str, Enum):
    """'top' or 'bottom'."""

    top = "top"
    bottom = "bottom"


class FilterKgraphTopNParameters(BaseModel):
    edge_attribute: str
    max_edges: Annotated[int, Field(le=0)] = 50
    remove_top_or_bottom: TopOrBottomEnum = TopOrBottomEnum.top
    qedge_keys: Optional[HashableSequence[str]] = None
    qnode_keys: HashableSequence[str] = Field(
        default_factory=lambda: HashableSequence[str]()
    )


class FilterKgraphPercentileParameters(BaseModel):
    edge_attribute: str
    threshold: Annotated[float, Field(ge=0, le=100)] = 95
    remove_above_or_below: AboveOrBelowEnum = AboveOrBelowEnum.below
    qedge_keys: Optional[HashableSequence[str]] = None
    qnode_keys: HashableSequence[str] = Field(
        default_factory=lambda: HashableSequence[str]()
    )


class OperationFilterKgraphPercentile(BaseOperation):
    id: Literal["filter_kgraph_percentile"]
    parameters: FilterKgraphPercentileParameters
    model_config = ConfigDict(extra="forbid")


class PlusOrMinusEnum(str, Enum):
    """'plus' or 'minus'."""

    plus = "plus"
    minus = "minus"


class FilterKgraphStdDevParameters(BaseModel):
    edge_attribute: str
    plus_or_minus_std_dev: PlusOrMinusEnum = PlusOrMinusEnum.plus
    num_sigma: Annotated[float, Field(ge=0)] = 1
    remove_above_or_below: AboveOrBelowEnum = AboveOrBelowEnum.below
    qedge_keys: Optional[HashableSequence[str]] = None
    qnode_keys: HashableSequence[str] = Field(
        default_factory=lambda: HashableSequence[str]()
    )


class OperationFilterKgraphStdDev(BaseOperation):
    id: Literal["filter_kgraph_std_dev"]
    parameters: FilterKgraphStdDevParameters
    model_config = ConfigDict(extra="forbid")


class OperationFilterKgraphTopN(BaseOperation):
    id: Literal["filter_kgraph_top_n"]
    parameters: FilterKgraphTopNParameters
    model_config = ConfigDict(extra="forbid")


class OperationLookup(BaseOperation):
    id: Literal["lookup"]
    parameters: Optional[Any] = None
    model_config = ConfigDict(extra="forbid")


class OperationLookupAndScore(BaseOperation):
    id: Literal["lookup_and_score"]
    parameters: Optional[Any] = None
    model_config = ConfigDict(extra="forbid")


class OperationOverlay(BaseOperation):
    id: Literal["overlay"]
    parameters: Optional[Any] = None
    model_config = ConfigDict(extra="forbid")


class OverlayComputeJaccardParameters(BaseModel):
    intermediate_node_key: str
    end_node_keys: HashableSequence[str]
    virtual_relation_label: str


class OperationOverlayComputeJaccard(BaseOperation):
    id: Literal["overlay_compute_jaccard"]
    parameters: OverlayComputeJaccardParameters
    model_config = ConfigDict(extra="forbid")


class OverlayComputeNgdParameters(BaseModel):
    qnode_keys: HashableSequence[str]
    virtual_relation_label: str


class OperationOverlayComputeNgd(BaseOperation):
    id: Literal["overlay_compute_ngd"]
    parameters: OverlayComputeNgdParameters
    model_config = ConfigDict(extra="forbid")


class OperationOverlayConnectKnodes(BaseOperation):
    id: Literal["overlay_connect_knodes"]
    parameters: Optional[Any] = None
    model_config = ConfigDict(extra="forbid")


class OverlayFisherExactTestParameters(BaseModel):
    subject_qnode_key: str
    object_qnode_key: str
    virtual_relation_label: str
    rel_edge_key: Optional[str] = None


class OperationOverlayFisherExactTest(BaseOperation):
    id: Literal["overlay_fisher_exact_test"]
    parameters: OverlayFisherExactTestParameters
    model_config = ConfigDict(extra="forbid")


class OperationRestate(BaseOperation):
    id: Literal["restate"]
    parameters: Optional[Any] = None
    model_config = ConfigDict(extra="forbid")


class OperationScore(BaseOperation):
    id: Literal["score"]
    parameters: Optional[Any] = None
    model_config = ConfigDict(extra="forbid")


class OperationSortResults(BaseOperation):
    id: Literal["sort_results"]
    parameters: Optional[Any] = None
    model_config = ConfigDict(extra="forbid")


class AscOrDescEnum(str, Enum):
    """'ascending' or 'descending'."""

    ascending = "ascending"
    descending = "descending"


class SortResultsEdgeAttributeParameters(BaseModel):
    edge_attribute: str
    ascending_or_descending: AscOrDescEnum
    qedge_keys: Optional[HashableSequence[str]] = None


class OperationSortResultsEdgeAttribute(BaseOperation):
    id: Literal["sort_results_edge_attribute"]
    parameters: SortResultsEdgeAttributeParameters
    model_config = ConfigDict(extra="forbid")


class SortResultsNodeAttributeParameters(BaseModel):
    node_attribute: str
    ascending_or_descending: AscOrDescEnum
    qnode_keys: Optional[HashableSequence[str]] = None


class OperationSortResultsNodeAttribute(BaseOperation):
    id: Literal["sort_results_node_attribute"]
    parameters: SortResultsNodeAttributeParameters
    model_config = ConfigDict(extra="forbid")


class SortResultsScoreParameters(BaseModel):
    ascending_or_descending: AscOrDescEnum


class OperationSortResultsScore(BaseOperation):
    id: Literal["sort_results_score"]
    parameters: SortResultsScoreParameters
    model_config = ConfigDict(extra="forbid")


operations = [
    OperationAnnotate,
    OperationAnnotateEdges,
    OperationAnnotateNodes,
    OperationBind,
    OperationCompleteResults,
    OperationEnrichResults,
    OperationFill,
    OperationFilterKgraph,
    OperationFilterKgraphContinuousKedgeAttribute,
    OperationFilterKgraphDiscreteKedgeAttribute,
    OperationFilterKgraphDiscreteKnodeAttribute,
    OperationFilterKgraphOrphans,
    OperationFilterKgraphPercentile,
    OperationFilterKgraphStdDev,
    OperationFilterKgraphTopN,
    OperationFilterResults,
    OperationFilterResultsTopN,
    OperationLookup,
    OperationLookupAndScore,
    OperationOverlay,
    OperationOverlayComputeJaccard,
    OperationOverlayComputeNgd,
    OperationOverlayConnectKnodes,
    OperationOverlayFisherExactTest,
    OperationRestate,
    OperationScore,
    OperationSortResults,
    OperationSortResultsEdgeAttribute,
    OperationSortResultsNodeAttribute,
    OperationSortResultsScore,
]

Operation = RootModel[
    Union[
        OperationAnnotate,
        OperationAnnotateEdges,
        OperationAnnotateNodes,
        OperationBind,
        OperationCompleteResults,
        OperationEnrichResults,
        OperationFill,
        OperationFilterKgraph,
        OperationFilterKgraphContinuousKedgeAttribute,
        OperationFilterKgraphDiscreteKedgeAttribute,
        OperationFilterKgraphDiscreteKnodeAttribute,
        OperationFilterKgraphOrphans,
        OperationFilterKgraphPercentile,
        OperationFilterKgraphStdDev,
        OperationFilterKgraphTopN,
        OperationFilterResults,
        OperationFilterResultsTopN,
        OperationLookup,
        OperationLookupAndScore,
        OperationOverlay,
        OperationOverlayComputeJaccard,
        OperationOverlayComputeNgd,
        OperationOverlayConnectKnodes,
        OperationOverlayFisherExactTest,
        OperationRestate,
        OperationScore,
        OperationSortResults,
        OperationSortResultsEdgeAttribute,
        OperationSortResultsNodeAttribute,
        OperationSortResultsScore,
    ]
]


Workflow = HashableSequence[Operation]



"""Setup file for reasoner package."""

from setuptools import setup

from pathlib import Path

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name="reasoner-pydantic",
    version="6.0.0",
    author="Abrar Mesbah",
    author_email="amesbah@covar.com",
    url="https://github.com/TranslatorSRI/reasoner-pydantic",
    description="Pydantic models for the Reasoner API data formats",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=["reasoner_pydantic"],
    include_package_data=True,
    install_requires=["pydantic>=2,<3"],
    zip_safe=False,
    license="MIT",
    python_requires=">=3.9",
)



import collections.abc
from typing import Any, Collection, Generic, Iterable, Optional, TypeVar, cast

from pydantic import RootModel, model_serializer

KeyType = TypeVar("KeyType")
ValueType = TypeVar("ValueType")


class HashableMapping(
    RootModel[dict[KeyType, ValueType]],
    collections.abc.MutableMapping[KeyType, ValueType],
    Generic[KeyType, ValueType],
):
    """
    Custom class that implements MutableMapping and is hashable
    """

    root: dict[KeyType, ValueType] = dict()

    def __getitem__(self, k: KeyType) -> ValueType:
        return self.root[k]

    def __iter__(self):
        return iter(self.root)

    def __len__(self) -> int:
        return len(self.root)

    def __setitem__(self, k: KeyType, v: ValueType) -> None:
        self.root[k] = v

    def __delitem__(self, k: KeyType) -> None:
        del self.root[k]

    def __hash__(self):
        return hash(tuple((k, v) for k, v in self.root.items()))

    def __eq__(self, other: object) -> bool:
        return self.__hash__() == other.__hash__()


class HashableSequence(
    RootModel[list[ValueType]],
    collections.abc.MutableSequence[ValueType],
    Generic[ValueType],
):
    """
    Custom class that implements MutableSequence and is hashable
    """

    root: list[ValueType] = list()

    def __contains__(self, v: object) -> bool:
        return v in self.root

    def __getitem__(self, i):
        return self.root[i]

    def __iter__(self):
        return iter(self.root)

    def __len__(self):
        return len(self.root)

    def __setitem__(self, i, v) -> None:
        self.root[i] = v

    def __delitem__(self, i):
        del self.root[i]

    def insert(self, index, value):
        self.root.insert(index, value)

    def __hash__(self):
        return hash(tuple(self.root))

    def __eq__(self, other: object) -> bool:
        return self.__hash__() == other.__hash__()


class HashableSet(
    RootModel[set[ValueType]],
    collections.abc.MutableSet[ValueType],
    Generic[ValueType],
):
    """
    Custom class that implements MutableSet and is hashable
    """

    root: set[ValueType] = set()

    def __contains__(self, v):
        return v in self.root

    def __iter__(self):
        return iter(self.root)

    def __len__(self):
        return len(self.root)

    def add(self, value):
        self.root.add(value)

    def update(self, other: Iterable[ValueType]):
        self.root.update(other)

    def discard(self, value):
        self.root.discard(value)

    def __hash__(self):
        # Use frozenset instead of tuple to ensure
        # hash is computed without ordering of elements
        return hash(frozenset(self.root))

    @model_serializer
    def as_list(self) -> list[ValueType]:
        """Custom serialization method to convert to list"""

        # Normally, the dict method tries to cast to the root type.
        # This isn't an issue for most root types, but here that causes:
        # set({"hello" : "world"}) which doesn't work because dicts are not hashable
        # This overrides that functionality to cast to a list instead
        return list(self.root)


def nonzero_validator(v: Optional[Collection[Any]]):
    if v is not None and len(v) == 0:
        raise ValueError("Must have nonzero number of elements")
    return v


def make_hashable(o: object):
    """
    Convert a generic Python object to a hashable one recursively

    This is an expensive operation, so it is best used sparingly
    """

    # type(o) is faster than isinstance(o) because it doesn't
    # traverse the inheritance hierarchy
    o_type = str(type(o))

    if "dict" in o_type:
        return HashableMapping(
            {k: make_hashable(v) for k, v in cast(dict[Any, Any], o).items()}
        )
    if "list" in o_type:
        return HashableSequence([make_hashable(v) for v in cast(list[Any], o)])

    return o



from typing import Any, Generic, TypeVar

from pydantic import (
    model_validator,
    BaseModel as PydanticBaseModel,
    RootModel as PydanticRootModel,
)

from .utils import make_hashable


class BaseModel(PydanticBaseModel):
    """
    Custom base model for all classes

    This provides hash and equality methods.
    """

    def __hash__(self) -> int:
        """Hash function based on Pydantic implementation"""
        # Faster than calling tuple() or otherwise unpacking dictionaries
        return hash(
            (
                self.__class__.__name__,
                *self.__dict__.items(),
                *((self.model_extra or {}).items()),
            )
        )

    def __eq__(self, other: object) -> bool:
        """Equality function that calls hash function"""
        return self.__hash__() == other.__hash__()

    def update(self, _other: object) -> None:
        """Update fields on this object with fields from other object"""
        raise NotImplementedError(
            f"Model {self.__class__.__name__} has no update method"
        )

    def get_field(self, field: str):
        return getattr(self, field, None)

    @model_validator(mode="after")
    def ensure_hashable_extra(self) -> "BaseModel":
        """Ensure extra fields are hashable, if they're allowed."""
        if not self.model_extra:
            return self
        for k, v in self.model_extra.items():
            self.model_extra[k] = make_hashable(v)
        return self

    def to_dict(self) -> dict[Any, Any]:
        """DEPRECATED: use model_dump() instead."""
        return self.model_dump()


RootType = TypeVar("RootType")


class RootModel(BaseModel, PydanticRootModel[RootType], Generic[RootType]):
    """
    Custom root model for all classes.

    This provides hash and equality methods.
    """



"""Query graph models."""

from enum import Enum

from typing import Annotated, Any, Optional

from pydantic import AfterValidator, ConfigDict, Field, model_serializer

from .base_model import BaseModel
from .utils import HashableMapping, HashableSequence, nonzero_validator
from .shared import BiolinkEntity, BiolinkPredicate, CURIE, KnowledgeType, Qualifier


class Operator(str, Enum):
    """Operator."""

    equal_to = "=="
    deep_equal_to = "==="
    greater_than = ">"
    less_than = "<"
    matches = "matches"


class AttributeConstraint(BaseModel):
    """QNode or QEdge attribute constraint."""

    name: str
    id: CURIE
    negated: Annotated[
        bool,
        Field(
            False,
            title="not",
            alias="not",
        ),
    ]
    operator: Operator
    value: Any
    unit_id: Optional[Any] = None
    unit_name: Optional[Any] = None
    model_config = ConfigDict(extra="forbid")

    @model_serializer
    def serialize(self):
        """Replace `negated` with `not`."""
        output = self.__dict__
        output["not"] = output.pop("negated", False)
        return output


class QualifierConstraint(BaseModel):
    """QEdge Qualifier constraint."""

    qualifier_set: HashableSequence[Qualifier] = Field(
        title="qualifier set", default_factory=lambda: HashableSequence[Qualifier]()
    )


class PathConstraint(BaseModel):
    """QPath Constraint."""

    intermediate_categories: Annotated[
        Optional[HashableSequence[BiolinkEntity]], AfterValidator(nonzero_validator)
    ]


class SetInterpretationEnum(str, Enum):
    """Enumeration for set interpretation."""

    BATCH = "BATCH"
    ALL = "ALL"
    MANY = "MANY"


class QNode(BaseModel):
    """Query node."""

    ids: Annotated[
        Optional[HashableSequence[CURIE]], AfterValidator(nonzero_validator)
    ] = None
    categories: Annotated[
        Optional[HashableSequence[BiolinkEntity]], AfterValidator(nonzero_validator)
    ] = None
    set_interpretation: Annotated[
        Optional[SetInterpretationEnum], Field(SetInterpretationEnum.BATCH)
    ]
    constraints: HashableSequence[AttributeConstraint] = Field(
        title="attribute constraints",
        default_factory=lambda: HashableSequence[AttributeConstraint](),
    )
    member_ids: Optional[HashableSequence[CURIE]] = Field(
        title="set member ids", default_factory=lambda: HashableSequence[CURIE]()
    )

    model_config = ConfigDict(
        title="query-graph node",
        extra="allow",
        populate_by_name=True,
        use_enum_values=True,  # See https://github.com/pydantic/pydantic/discussions/9270
        validate_default=True,  # See https://docs.pydantic.dev/latest/api/config/#pydantic.config.ConfigDict.use_enum_values
    )


class QEdge(BaseModel):
    """Query edge."""

    subject: Annotated[
        str,
        Field(
            title="subject node id",
        ),
    ]
    object: Annotated[
        str,
        Field(
            title="object node id",
        ),
    ]
    knowledge_type: Annotated[
        Optional[KnowledgeType], Field(title="knowledge type")
    ] = None
    predicates: Annotated[
        Optional[HashableSequence[BiolinkPredicate]], AfterValidator(nonzero_validator)
    ] = None
    attribute_constraints: HashableSequence[AttributeConstraint] = Field(
        title="attribute constraints",
        default_factory=lambda: HashableSequence[AttributeConstraint](),
    )
    qualifier_constraints: HashableSequence[QualifierConstraint] = Field(
        title="qualifier constraint",
        default_factory=lambda: HashableSequence[QualifierConstraint](),
    )

    model_config = ConfigDict(
        title="query-graph edge", extra="allow", populate_by_name=True
    )


class QPath(BaseModel):
    """Query path."""

    subject: Annotated[
        str,
        Field(
            title="subject node id",
        ),
    ]

    object: Annotated[
        str,
        Field(
            title="object node id",
        ),
    ]

    predicates: Annotated[
        Optional[HashableSequence[BiolinkPredicate]],
        Field(
            title="predicates",
        ),
    ] = None

    constraints: Optional[HashableSequence[PathConstraint]]


class BaseQueryGraph(BaseModel):
    """Base query graph."""

    nodes: Annotated[
        HashableMapping[str, QNode],
        Field(
            title="dict of nodes",
        ),
    ]

    model_config = ConfigDict(title="simple query graph", extra="allow")


class QueryGraph(BaseQueryGraph):
    """Traditional query graph."""

    edges: Annotated[
        HashableMapping[str, QEdge],
        Field(
            title="dict of edges",
        ),
    ]

    model_config = ConfigDict(title="simple query graph", extra="allow")


class PathfinderQueryGraph(BaseQueryGraph):
    """Pathfinder query graph."""

    paths: Annotated[
        HashableMapping[str, QPath],
        Field(
            title="dict of paths",
        ),
    ]

    model_config = ConfigDict(title="pathfinder query graph", extra="allow")
